package no.ntnu.mycbr.rest;

import de.dfki.mycbr.core.DefaultCaseBase;
import de.dfki.mycbr.core.Project;
import de.dfki.mycbr.core.casebase.*;
import de.dfki.mycbr.core.casebase.Attribute;
import de.dfki.mycbr.core.model.*;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;

import java.io.*;
import java.util.*;

class AddCaseAndRecipe {
    private String fileName = "./src/main/resources/sandwich-recipes.json";

    private String returnMessage;
    private String casebaseFailureMessage = "casebase";
    private String recipeFailureMessage = "recipe";
    private String successMessage = "success";

    AddCaseAndRecipe(String caseBase, String concept, String recipe, HashMap<String, Object> caseAttributes) {

        Project project = App.getProject();
        DefaultCaseBase cb = (DefaultCaseBase) project.getCaseBases().get(caseBase);
        Concept myConcept = project.getConceptByID(concept);

        // Remove "-Adapted" extension from caseName if it is there (it should be removed already)
        String caseName = caseAttributes.get("CaseName").toString().replace("-Adapted", "");
        caseAttributes.remove("CaseName");

        // Check whether the caseName is already taken, create extension for the new caseInstance
        Integer extension = getCaseExtension(myConcept, caseName, 1);

        // Get current id from the recipe string
        JSONParser parser = new JSONParser();
        String recipeId = "";
        try {
            Object obj = parser.parse(recipe);
            JSONObject jsonObject = (JSONObject) obj;
            recipeId = jsonObject.get("id").toString();
        } catch (Exception e) {
            System.out.println("Failed to add case recipe to file " + fileName + " while parsing string to JSON");
            this.returnMessage = recipeFailureMessage;
        }

        Instance caseInstance = createInstanceFromAttributeList(myConcept, caseAttributes, caseName + "-" + extension);
        caseInstance.setAdapted(true);

        try {
            cb.addCase(caseInstance);

            // Replace "oldId" with "newId" and add recipe to JSON
            if (addRecipe(recipe.replace("\"" + recipeId + "\"", "\"" + recipeId + "-" + extension + "\""))) {
                this.returnMessage = successMessage;
                System.out.println("Added case \"" + caseInstance.getName() + "\" to casebase " + cb.getName());
            } else {
                cb.removeCase(caseInstance);
                this.returnMessage = recipeFailureMessage;
                System.out.println("Failed to add case recipe to file " + fileName);
            }

            project.save();
        } catch (Exception e) {
            this.returnMessage = casebaseFailureMessage;
            System.out.println("Failed to add case \"" + caseInstance.getName() + "\" to casebase " + cb.getName());
        }
    }

    private Instance createInstanceFromAttributeList(Concept myConcept, HashMap<String, Object> attributeList, String caseName) {
        if (attributeList != null) {
            try {
                Instance caseInstance = new Instance(myConcept, caseName);
                for (Map.Entry<String, Object> attributeValueSet : attributeList.entrySet()) {
                    String attributeName = attributeValueSet.getKey();
                    AttributeDesc attributeDesc = myConcept.getAttributeDesc(attributeName);
                    if ("FloatDesc".equalsIgnoreCase(attributeDesc.getClass().getSimpleName())) {
                        caseInstance.addAttribute(attributeDesc, Float.parseFloat(attributeValueSet.getValue().toString()));
                    }
                    if ("IntegerDesc".equalsIgnoreCase(attributeDesc.getClass().getSimpleName())) {
                        caseInstance.addAttribute(attributeDesc, Integer.parseInt(attributeValueSet.getValue().toString()));
                    }
                    if ("DoubleDesc".equalsIgnoreCase(attributeDesc.getClass().getSimpleName())) {
                        caseInstance.addAttribute(attributeDesc, Double.parseDouble(attributeValueSet.getValue().toString()));
                    }

                    if ("StringDesc".equalsIgnoreCase(attributeDesc.getClass().getSimpleName())) {
                        caseInstance.addAttribute(attributeDesc, attributeValueSet.getValue().toString());
                    }
                    if ("SymbolDesc".equalsIgnoreCase(attributeDesc.getClass().getSimpleName())) {
                        SymbolDesc symbolAttribute = (SymbolDesc) attributeDesc;
                        if (!symbolAttribute.isMultiple()) {
                            caseInstance.addAttribute(attributeDesc, attributeValueSet.getValue());
                        } else {
                            LinkedList<Attribute> attributeLinkedList = new LinkedList<>();
                            StringTokenizer stringTokenizer = new StringTokenizer((String) attributeValueSet.getValue(), ";");
                            while (stringTokenizer.hasMoreElements()) {
                                String symbolName = stringTokenizer.nextElement().toString().trim();
                                attributeLinkedList.add(symbolAttribute.getAttribute(symbolName));
                            }
                            MultipleAttribute<SymbolDesc> multiSymbol = new MultipleAttribute<>(symbolAttribute, attributeLinkedList);
                            caseInstance.addAttribute(attributeDesc, multiSymbol);
                        }
                    }
                }
                return caseInstance;

            } catch (Exception e) {
                this.returnMessage = casebaseFailureMessage;
            }
        }
        return new Instance(myConcept, "ERROR-INSTANCE");
    }

    private Integer getCaseExtension(Concept myConcept, String caseName, Integer extension) {
        Instance alreadyExistingCaseWithName = myConcept.getInstance(caseName + "-" + extension);
        if (alreadyExistingCaseWithName == null) {
            return extension;
        }
        return getCaseExtension(myConcept, caseName, extension + 1);
    }

    private boolean addRecipe(String newRecipe) {
        List<String> lines = new ArrayList<>();

        // Read the current file into lines
        BufferedReader r = null;
        try {
            r = new BufferedReader(new FileReader(fileName));
            String in;
            while ((in = r.readLine()) != null)
                lines.add(in);
        } catch (Exception e) {
            this.returnMessage = recipeFailureMessage;
            System.out.println("Failed to add recipe to " + fileName + " while trying to read file.");
            return false;
        } finally {
            try {
                r.close();
            } catch (Exception e) {
                this.returnMessage = recipeFailureMessage;
                System.out.println("Failed to add recipe to " + fileName + " while trying to read file.");
                return false;
            }
        }

        // This is done to make sure we dont fuck up the file
        String secondFromBottomLine = lines.get(lines.size() - 2);
        String bottomLine = lines.get(lines.size() - 1);
        if (secondFromBottomLine.replaceAll("\\s+", "").matches("]") && bottomLine.replaceAll("\\s+", "").matches("}")) {
            lines.remove(lines.size() - 1);
            lines.remove(lines.size() - 1);

            lines.add(",");
            lines.add(newRecipe);
            lines.add("]");
            lines.add("}");
        }

        // Write all lines back to the file and save
        PrintWriter w = null;
        try {
            w = new PrintWriter(new FileWriter(fileName));
            for (String line : lines)
                w.println(line);
            System.out.println("Recipe added to " + fileName);
        } catch (Exception e) {
            this.returnMessage = recipeFailureMessage;
            System.out.println("Failed to add recipe to " + fileName + " while trying to write file.");
            return false;
        } finally {
            try {
                w.close();
            } catch (Exception e) {
                this.returnMessage = recipeFailureMessage;
                System.out.println("Failed to add recipe to " + fileName + " while trying to write file.");
                return false;
            }
        }

        return true;
    }

    String addNewCase() {
        return returnMessage;
    }
}
