package no.ntnu.mycbr.rest.adaptation;
import de.dfki.mycbr.core.casebase.Instance;
import de.dfki.mycbr.core.model.Concept;
import de.dfki.mycbr.core.similarity.Similarity;
import de.dfki.mycbr.util.Pair;
import no.ntnu.mycbr.CBREngine;
import no.ntnu.mycbr.rest.NewRetrieval;
import no.ntnu.mycbr.rest.adaptation.adaptedInstance.AdaptationHistory;
import no.ntnu.mycbr.rest.adaptation.adaptedInstance.AdaptationProcess;
import no.ntnu.mycbr.rest.adaptation.adaptedInstance.SuitableAdaptation;
import no.ntnu.mycbr.rest.adaptation.adaptedInstance.TitleAdaptation;
import no.ntnu.mycbr.rest.adaptation.rules.AdaptationSubstitution;

import java.util.*;


// Runs AdaptationProcess for each retrieved recipe (Query)
// Contains all statistics about all adapted cases e.g
// Prints history for adapted cases
public class Adaptation {
    private List<de.dfki.mycbr.core.casebase.Instance> modifiedValidInstances = new ArrayList<Instance>();

    private List<AdaptationProcess> adaptationProcesses = new ArrayList<AdaptationProcess>();
    private List<AdaptationProcess> invalidAdaptationProcesses = new ArrayList<AdaptationProcess>();
    private List<AdaptationProcess> zeroAdaptationProcesses = new ArrayList<AdaptationProcess>();
    private HashMap<Instance, List<Pair<String,String>>> instanceSubstitutions = new HashMap<Instance, List<Pair<String,String>>>();
    private HashMap<String, AdaptationHistory> instanceHistory = new HashMap<String, AdaptationHistory>();
    private AdaptationHistory globalStatistics = new AdaptationHistory();
    private Instance queryDesired;
    private Instance queryUndesired;
    private Concept myConcept;
    private List<Pair<Instance, Similarity>> punishedAdaptedResults;
    private List<Pair<Instance, Similarity>> filteredOriginalResults;

    public Adaptation(Instance queryDesired, Instance queryUndesired, List<Pair<Instance, Similarity>> originalResults) throws Exception {
        this.queryDesired = queryDesired;
        this.queryUndesired = queryUndesired;
        this.myConcept = CBREngine.getConcept();

        this.runAdaptation(originalResults);
        this.postAdaptationProcess(originalResults);
    }

    // runs the whole adaptation process
    // iterates every original instance and creates an adapted version
    private void runAdaptation(List<Pair<Instance, Similarity>> results) {
        for (Pair<Instance, Similarity> result : results) {
            Instance recipeInstance = result.getFirst();
            try {
                AdaptationProcess adaptedInstanceProcess = new AdaptationProcess(recipeInstance, queryDesired, queryUndesired);
                if (!adaptedInstanceProcess.getIsValid()) {
                    // add adaptation processes that are either not valid or not modified - to this list
                    invalidAdaptationProcesses.add(adaptedInstanceProcess);
                }

                else if (adaptedInstanceProcess.getIsModified()) {
                    // add to valid adaptation processes
                    adaptationProcesses.add(adaptedInstanceProcess);

                    // add to modified instances
                    modifiedValidInstances.add(adaptedInstanceProcess.getAdaptedInstance());

                    // add substitutions to <adaptedInstance,substitutions> overview
                    this.addSubstitutions(adaptedInstanceProcess, adaptedInstanceProcess.getAdaptedInstance());

                    // add statistics to overview hashmap
                    instanceHistory.put(adaptedInstanceProcess.getAdaptedInstance().getName(), adaptedInstanceProcess.getHistory());

                    // add instance statistics to global statistics
                    this.updateGlobalStatistics(adaptedInstanceProcess);
                }
                else if (!adaptedInstanceProcess.getHistory().desiredChanges() && !adaptedInstanceProcess.getHistory().undesiredChanges()){
                    // adaptation processes that are not modified, but valid = recipes where no modifications are made
                    zeroAdaptationProcesses.add(adaptedInstanceProcess);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    private void postAdaptationProcess(List<Pair<Instance, Similarity>> originalResults) {
        // Start retrieval for adapted instances
        List<Pair<Instance, Similarity>> adaptedResults = NewRetrieval.retrieveMultiple(queryDesired, modifiedValidInstances, myConcept);

        // Remove adapted recipes if some contains exactly the same ingredients. So that we don't switch back and forth with adaptations
        PostAdaptationAdapted.removeEqualRecipes(originalResults, adaptedResults);

        // Punish adapted instances
        punishedAdaptedResults = PostAdaptationAdapted.punishAdaptedInstances(originalResults, adaptedResults, myConcept, this);

        // Do suitable ingredients substitution
        SuitableAdaptation.doSuitableAdaptation(this, originalResults, adaptedResults, queryDesired, queryUndesired);

        TitleAdaptation.doTitleAdaptation(this, originalResults, adaptedResults, queryDesired, queryUndesired);

        // Print valid adapted processes
        PostAdaptation.pairAndPrintResults(this, originalResults, adaptedResults, punishedAdaptedResults, queryDesired);

        // Filter retrieved original instances (make sure they satisfy undesired query) - This must happen after printing! So that it is not removed :)
        filteredOriginalResults = PostAdaptationOriginal.filterOriginalRecipes(myConcept, queryUndesired, originalResults);
    }

    public List<Pair<Instance, Similarity>> getFinalAdaptedResults() {
        return punishedAdaptedResults;
    }

    public List<Pair<Instance, Similarity>> getFinalOriginalResults() {
        return filteredOriginalResults;
    }


    private void updateGlobalStatistics(AdaptationProcess adaptedInstanceProcess) {
        // desired statistics
        int aCount = adaptedInstanceProcess.getHistory().getDesiredAdditionRuleCount();
        globalStatistics.setDesiredAdditionRuleCount(globalStatistics.getDesiredAdditionRuleCount() + aCount);
        int sCount = adaptedInstanceProcess.getHistory().getDesiredSubstitutionRuleCount();
        globalStatistics.setDesiredSubstitutionRuleCount(globalStatistics.getDesiredSubstitutionRuleCount() + sCount);
        int ssCount = adaptedInstanceProcess.getHistory().getDesiredSimilaritySubstitutionCount();
        globalStatistics.setDesiredSimilaritySubstitutionCount(globalStatistics.getDesiredSimilaritySubstitutionCount() + ssCount);

        // undesired statistics
        int urCount = adaptedInstanceProcess.getHistory().getUndesiredRemoveRuleCount();
        globalStatistics.setUndesiredRemoveRuleCount(globalStatistics.getUndesiredRemoveRuleCount() + urCount);
        int usCount = adaptedInstanceProcess.getHistory().getUndesiredSubstitutionRuleCount();
        globalStatistics.setUndesiredSubstitutionRuleCount(globalStatistics.getUndesiredSubstitutionRuleCount() + usCount);
        int ussCount = adaptedInstanceProcess.getHistory().getUndesiredSimilaritySubstitutionCount();
        globalStatistics.setUndesiredSimilaritySubstitutionCount(globalStatistics.getUndesiredSimilaritySubstitutionCount() + ussCount);
    }

    private void addSubstitutions(AdaptationProcess adaptCase, Instance adaptedInstance) {
        List<Pair<String,String>> substitutions = new ArrayList<Pair<String,String>>();
        for (AdaptationSubstitution subst : adaptCase.getSubstitutions()) {
            substitutions.add(new Pair<String, String>(subst.getIngredient1(), subst.getIngredient2()));
        }
        instanceSubstitutions.put(adaptedInstance, substitutions);
    }

    public void addSuitableSubstitutions(Instance adaptedInstance, List<AdaptationSubstitution> substitutions) {
        for (AdaptationSubstitution subst : substitutions) {
            instanceSubstitutions.get(adaptedInstance).add(new Pair<String, String>(subst.getIngredient1(), subst.getIngredient2()));
        }
    }

    public List<AdaptationProcess> getAdaptationProcesses() {
        return adaptationProcesses;
    }

    public AdaptationProcess getAdaptationProcessForInstance(Instance adaptedInstance) {
        for (AdaptationProcess adaptationProcess : adaptationProcesses) {
            if (adaptationProcess.getAdaptedInstance().getName().equals(adaptedInstance.getName())) {
                return adaptationProcess;
            }
        }
        return null;
    }

    public List<AdaptationProcess> getInvalidAdaptationProcesses() {
        return invalidAdaptationProcesses;
    }

    public List<AdaptationProcess> getZeroAdaptationProcesses() {
        return zeroAdaptationProcesses;
    }

    public AdaptationHistory getGlobalStatistics() {
        return globalStatistics;
    }

    public List<Pair<String,String>> getSubstitutionsForInstance(Instance instance) {
        return instanceSubstitutions.get(instance);
    }

}