package no.ntnu.mycbr.rest.adaptation.adaptedInstance;

import de.dfki.mycbr.core.model.AttributeDesc;
import de.dfki.mycbr.core.casebase.Instance;
import no.ntnu.mycbr.CBREngine;
import no.ntnu.mycbr.rest.adaptation.helpClasses.Operations;
import no.ntnu.mycbr.rest.adaptation.rules.AdaptationSubstitution;
import de.dfki.mycbr.core.model.Concept;

import java.util.*;


public class AdaptationProcess {
    // Class that contains all information about the process of adapting an instance

    private Instance adaptedInstance;
    private Instance originalInstance;
    private int adaptationThreshold;
    private AdaptationHistory history;
    private List<AdaptationSubstitution> substitutions = new ArrayList<AdaptationSubstitution>();
    private Boolean recipeIsModified = false;
    private Boolean recipeIsValid = true;
    private Concept myConcept;

    public AdaptationProcess(Instance oInstance, Instance qDesired, Instance qUndesired) throws Exception {
        this.originalInstance = oInstance;
        this.adaptedInstance = this.setAsOriginalRecipe();
        this.myConcept = CBREngine.getConcept();

        // set adaptation threshold
        this.setAdaptationThreshold();

        // history
        history = new AdaptationHistory(oInstance, adaptedInstance);

        // modify adapted adaptedInstance according to undesired query
        UndesiredAdaptation adaptUndesired = new UndesiredAdaptation(originalInstance, adaptedInstance, qUndesired, history, adaptationThreshold, myConcept);

        // modify adapted adaptedInstance according to desired query (original instance is here the adapted instance from undesired)
        Instance instanceModifiedByUndesired = this.getCopy(adaptedInstance, new Instance(CBREngine.getConcept(), originalInstance.getName()+"-ModifiedByUndesired"));
        DesiredAdaptation adaptDesired = new DesiredAdaptation(instanceModifiedByUndesired, adaptedInstance, qDesired, qUndesired, history, adaptationThreshold, myConcept);

        // update information (valid, invalid, not modified) and gather all substitutions made
        this.updateModifiedInstanceInfo(adaptDesired, adaptUndesired);
    }

    private void updateModifiedInstanceInfo(DesiredAdaptation adaptDesired, UndesiredAdaptation adaptUndesired) {
        if (adaptDesired.modificationIsMade() || adaptUndesired.modificationIsMade()) {
            adaptedInstance.setAdapted(true);
            recipeIsModified = true;
        }
        for (AdaptationSubstitution subst : adaptDesired.getSubstitutions()) {
            substitutions.add(subst);
        }
        for (AdaptationSubstitution subst : adaptUndesired.getSubstitutions()) {
            substitutions.add(subst);
        }
        if (!adaptUndesired.undesiredIngredientsAreSatisfied()) {
            recipeIsValid = false;
        }
    }

    private void setAdaptationThreshold() {
        HashSet<String> allIngredientsFromInstance = Operations.getAllIngredientsFromInstance(originalInstance);
        adaptationThreshold = (int) Math.ceil(allIngredientsFromInstance.size());
    }

    private Instance setAsOriginalRecipe() {
        Instance adaptedInstance = new Instance(CBREngine.getConcept(), originalInstance.getName() + "-Adapted");
        HashMap<String, AttributeDesc> attributeDescs = CBREngine.getConcept().getAttributeDescs();
        for (Map.Entry<String, AttributeDesc> entry : attributeDescs.entrySet()) {
            AttributeDesc attrDesc = entry.getValue();
            adaptedInstance.addAttribute(attrDesc, originalInstance.getAttForDesc(attrDesc));
        }
        return adaptedInstance;
    }

    private Instance getCopy(Instance oldInstance, Instance newInstance) {
        HashMap<String, AttributeDesc> attributeDescs = CBREngine.getConcept().getAttributeDescs();
        for (Map.Entry<String, AttributeDesc> entry : attributeDescs.entrySet()) {
            AttributeDesc attrDesc = entry.getValue();
            newInstance.addAttribute(attrDesc, oldInstance.getAttForDesc(attrDesc));
        }
        return newInstance;
    }


    public AdaptationHistory getHistory() {
        return history;
    }
    public Instance getAdaptedInstance() {
        return adaptedInstance;
    }
    public Boolean getIsModified() {
        return recipeIsModified;
    }
    public Boolean getIsValid() {return recipeIsValid; }
    public List<AdaptationSubstitution> getSubstitutions() { return substitutions; }

}
