package no.ntnu.mycbr.rest.adaptation.adaptedInstance;

import de.dfki.mycbr.core.casebase.Attribute;
import de.dfki.mycbr.core.casebase.Instance;
import de.dfki.mycbr.core.model.*;
import no.ntnu.mycbr.rest.adaptation.helpClasses.Operations;
import no.ntnu.mycbr.rest.adaptation.helpClasses.OperationsTaxonomy;
import no.ntnu.mycbr.rest.adaptation.rules.AdaptationSubstitution;
import no.ntnu.mycbr.rest.adaptation.helpClasses.Settings;
import no.ntnu.mycbr.rest.adaptation.helpClasses.SubstitutionMethods;

import java.util.*;


public class DesiredAdaptation {
    // Adaptation with the desired query

    private Instance queryDesired;
    private Instance originalInstance;
    private Instance adaptedInstance;
    private int adaptationThreshold;
    private Boolean modificationIsMade = false;
    private Concept myConcept;
    private AdaptationHistory history;
    private List<AdaptationSubstitution> adaptationSubstitutions = new ArrayList<AdaptationSubstitution>();
    private HashSet<String> allCurrentIngredients = new HashSet<String>();

    public DesiredAdaptation(Instance oInstance, Instance aInstance, Instance qDesired, Instance qUndesired, AdaptationHistory history, int adaptationThreshold, Concept myConcept) throws Exception {
        this.queryDesired = qDesired;
        this.originalInstance = oInstance;
        this.adaptedInstance = aInstance;
        this.myConcept = myConcept;
        this.history = history;
        this.adaptationThreshold = adaptationThreshold;
        this.allCurrentIngredients = Operations.getAllIngredientsFromInstance(oInstance);
        this.runDesiredAdaptation();
    }

    private void runDesiredAdaptation() throws Exception {
        history.addLine("");
        history.addLine("");
        history.addLine("--------- Adapt according to desired query ---------");
        history.addLine("");

        // if there are ingredient requirements on queryDesired, loop through each AttributeDesc
        if (this.requirements(queryDesired)) {
            this.adaptEachAttribute();
        }
        else {
            history.addLine("There are no desired requirements in this query");
        }
    }

    protected Boolean requirements(Instance query) {
        for (Map.Entry<String, AttributeDesc> entry : myConcept.getAttributeDescs().entrySet()) {
            Attribute queryAttribute = query.getAttForDesc(entry.getValue());
            if (Operations.attributeContainsValues(queryAttribute)) {
                return true;
            }
        }
        return false;
    }

    // goes through each attributeDesc(Basic,Vegetable,etc..) and modifies if necessary according to query
    // returns adapted adaptedInstance
    private void adaptEachAttribute() throws Exception {
        for (Map.Entry<String, AttributeDesc> entry : myConcept.getAttributeDescs().entrySet()) {
            AttributeDesc attrDesc = entry.getValue();
            Attribute queryAttribute = queryDesired.getAttForDesc(attrDesc);
            Attribute recipeAttribute = originalInstance.getAttForDesc(attrDesc);
            if (!Settings.getAttributesThatShouldNotBeModified().contains(attrDesc.getName()) && Operations.attributeContainsValues(queryAttribute)) {
                String modifiedIngredients = this.getModifiedAttributeAsString(recipeAttribute, queryAttribute, attrDesc);
                Operations.modifyAttrDesc(attrDesc, modifiedIngredients, history, adaptedInstance);
            }
        }
    }

    private String getModifiedAttributeAsString(Attribute recipeAttribute, Attribute queryAttribute, AttributeDesc attrDesc) throws Exception {
        List<String> resultValues = Operations.getAttributeAsSingleValues(recipeAttribute);
        List<String> queryValues = Operations.getAttributeAsSingleValues(queryAttribute);

        // Either add or substitute from original ingredients. Or do nothing if no rules apply
        for (String desiredIngredient : queryValues) {
            // Updates current ingredients by adding ingredients of all other attributes, and this current updated attribute

            allCurrentIngredients = Operations.updateCurrentIngredients(attrDesc, resultValues, adaptedInstance);
            history.addLine("");
            history.addLine("- Required desired ingredient: " + desiredIngredient + ", considered with: " + allCurrentIngredients);
            if (resultValues.contains(desiredIngredient)) {
                history.addLine(".. Result already contains " + desiredIngredient);
                continue;
            }
            else {
                // If recipe contains a child (cherry tomato) of desired ingredient (tomato)
                // or if recipe contains a parent (tomato) of desired ingredient (cherry tomato)
                if (this.containsChildOfDesiredIngredient(desiredIngredient) || this.containsParentOfDesiredIngredient(desiredIngredient)) {
                    continue;
                }
            }

            if (history.getAdaptationPoints() >= adaptationThreshold) {
                history.addLine(".. Adaptation threshold: " + adaptationThreshold + " is reached. Can not consider: " + desiredIngredient);
                continue;
            }

            // Try modifying
            if (Operations.ingredientIsValid(desiredIngredient)) {
                AdaptationSubstitution subst;

                // consider simply adding rules
                subst = SubstitutionMethods.considerSimpleAdditionRules(resultValues, desiredIngredient, allCurrentIngredients, history);
                if (subst != null) {
                    modificationIsMade = true;
                    adaptationSubstitutions.add(subst);
                    continue;
                }
                // consider substitution rules
                subst = SubstitutionMethods.considerDesiredSubstitutionRules(queryAttribute, resultValues, desiredIngredient, allCurrentIngredients, history);
                if (subst != null) {
                    modificationIsMade = true;
                    adaptationSubstitutions.add(subst);
                    // update offer attribute
                    String offer = subst.getIngredient1();
                    AttributeDesc offerAttrDesc = Operations.findAttrDescForValue(offer);
                    Attribute offerAttribute = adaptedInstance.getAttForDesc(offerAttrDesc);
                    List<String> offerAttributeValues = Operations.getAttributeAsSingleValues(offerAttribute);
                    offerAttributeValues.remove(offer);
                    String ModifiedAttributeOfferAsString = Operations.getListAsString(offerAttributeValues);
                    Operations.modifyAttrDesc(offerAttrDesc, ModifiedAttributeOfferAsString, history, adaptedInstance);
                    continue;
                }
                // consider similarity substitution
                try {
                    AdaptationSubstitution substitution = SubstitutionMethods.considerSimilaritySubstituion(attrDesc, desiredIngredient, resultValues, queryDesired, history);
                    if (substitution != null) {
                        modificationIsMade = true;
                        adaptationSubstitutions.add(substitution);
                    }
                } catch (Exception e) {
                    System.out.println(e);
                    System.out.println("Error in considering similarity substitution");
                }
            }
            else {
                history.addLine("..QueryValue is invalid");
            }
        }

        // If no values are added to resultList and it was previously unknown, return null (then it will be set to unknown again)
        return Operations.getListAsString(resultValues);
    }



    private Boolean containsChildOfDesiredIngredient(String desiredIngredient) {
        try {
            List<String> childrenOfIngredient = OperationsTaxonomy.getChildrenOfIngredient(desiredIngredient);
            //history.addLine("children of ingredient " + desiredIngredient + ": " + childrenOfIngredient.toString());
            for (String child : childrenOfIngredient) {
                if (allCurrentIngredients.contains(child)) {
                    history.addLine("Result already contains " + child + ", which is child of " + desiredIngredient);
                    return true;
                }
            }
        } catch(Exception e) {
            System.out.println("Error in desired adaptation - containsChildOfDesiredIngredient");
            e.printStackTrace();
        }
        return false;
    }

    private Boolean containsParentOfDesiredIngredient(String desiredIngredient) {
        try {
            //HashSet<String> childrenOfAllCurrentIngredients = Operations.getChildrenOfSeveralIngredients(allCurrentIngredients, myConcept);
            //history.addLine("children of all current ingredients: " + childrenOfAllCurrentIngredients);
            for (String parent : allCurrentIngredients) {
                List<String> childrenOfParent = OperationsTaxonomy.getChildrenOfIngredient(parent);
                for (String child : childrenOfParent) {
                    if (desiredIngredient.equals(child)) {
                        history.addLine("Result already contains parent of " + desiredIngredient + "; " + parent);
                        return true;
                    }
                }
            }
        } catch(Exception e) {
            System.out.println("Error in desired adaptation - containsParentOfDesiredIngredient");
            e.printStackTrace();
        }
        return false;
    }

    public Boolean modificationIsMade() { return modificationIsMade; }
    public List<AdaptationSubstitution> getSubstitutions() { return adaptationSubstitutions; }

}
