package no.ntnu.mycbr.rest.adaptation.adaptedInstance;

import de.dfki.mycbr.core.casebase.Attribute;
import de.dfki.mycbr.core.casebase.Instance;
import de.dfki.mycbr.core.model.AttributeDesc;
import de.dfki.mycbr.core.similarity.*;
import no.ntnu.mycbr.CBREngine;
import no.ntnu.mycbr.rest.adaptation.helpClasses.Operations;
import no.ntnu.mycbr.rest.adaptation.helpClasses.Settings;
import no.ntnu.mycbr.rest.adaptation.rules.AdaptationSubstitution;
import de.dfki.mycbr.core.model.Concept;
import no.ntnu.mycbr.rest.adaptation.helpClasses.SubstitutionMethods;

import java.util.*;

public class UndesiredAdaptation {
    // Adaptation with undesired query

    private Instance queryUndesired;
    private Boolean undesiredIngredientsAreSatisfied = true;
    private Instance originalInstance;
    private Instance adaptedInstance;
    private int adaptationThreshold;
    private Boolean modificationIsMade = false;
    private Concept myConcept;
    private AdaptationHistory history;
    private List<AdaptationSubstitution> adaptationSubstitutions = new ArrayList<AdaptationSubstitution>();
    private HashSet<String> allCurrentIngredients = new HashSet<String>();

    public UndesiredAdaptation(Instance oInstance, Instance aInstance, Instance qUndesired, AdaptationHistory history, int adaptationThreshold, Concept myConcept) {
        this.queryUndesired = qUndesired;
        this.originalInstance = oInstance;
        this.adaptedInstance = aInstance;
        this.myConcept = myConcept;
        this.history = history;
        this.adaptationThreshold = adaptationThreshold;
        this.allCurrentIngredients = Operations.getAllIngredientsFromInstance(oInstance);
        this.runUndesiredAdaptation();
    }

    private void runUndesiredAdaptation() {
        history.addLine("");
        history.addLine("");
        history.addLine("--------- Adapt according to undesired query ---------");
        history.addLine("");

        // if there are ingredient requirements on queryUndesired, loop through each AttributeDesc
        if (this.requirements(queryUndesired)) {
            this.adaptEachAttribute();
        }
        else {
            history.addLine("There are no undesired requirements in this query");
        }
    }

    protected Boolean requirements(Instance query) {
        for (Map.Entry<String, AttributeDesc> entry : myConcept.getAttributeDescs().entrySet()) {
            Attribute queryAttribute = query.getAttForDesc(entry.getValue());
            if (Operations.attributeContainsValues(queryAttribute)) {
                return true;
            }
        }
        return false;
    }

    // goes through each attributeDesc(Basic,Vegetable,etc..) and modifies if necessary according to query
    // returns adapted adaptedInstance
    private void adaptEachAttribute() {
        for (Map.Entry<String, AttributeDesc> entry : myConcept.getAttributeDescs().entrySet()) {
            AttributeDesc attrDesc = entry.getValue();
            Attribute queryAttribute = queryUndesired.getAttForDesc(attrDesc);
            Attribute recipeAttribute = originalInstance.getAttForDesc(attrDesc);
            if (Operations.attributeContainsValues(queryAttribute) && !Settings.getAttributesThatShouldNotBeModified().contains(attrDesc.getName())) {
                history.addLine("");
                history.addLine(attrDesc.getName() + "/" + attrDesc);
                history.addLine("..Tries modifying");
                try {
                    String modifiedIngredients = this.getModifiedAttributeAsStringUndesired(recipeAttribute, queryAttribute, attrDesc);
                    Operations.modifyAttrDesc(attrDesc, modifiedIngredients, history, adaptedInstance);
                } catch (Exception e) {
                    System.out.println(e);
                    System.out.println("Error trying to fetch modifiedAttributeAsStringUndesired (prob. because of taxonomy)");
                }
            }
            // else: Skip modification.
        }
    }

    private String getModifiedAttributeAsStringUndesired(Attribute recipeAttribute, Attribute queryAttribute, AttributeDesc attrDesc) throws Exception {
        List<String> resultValues = Operations.getAttributeAsSingleValues(recipeAttribute);
        List<String> queryValues = Operations.getAttributeAsSingleValues(queryAttribute);
        Boolean simpleRemoveRuleFound;
        Boolean substitutionRuleFound;
        Boolean similaritySubstitutionRuleFound;

        for (String undesiredIngredient : queryValues) {
            // Updates current ingredients by adding ingredients of all other attributes, and this current updated attribute
            Operations.updateCurrentIngredients(attrDesc, resultValues,adaptedInstance);

            simpleRemoveRuleFound = false;
            substitutionRuleFound = false;
            similaritySubstitutionRuleFound = false;

            history.addLine("");
            history.addLine("- Required undesired ingredient: " + undesiredIngredient);
            history.addLine("..Considers with ingredients: " + allCurrentIngredients);
            if (!resultValues.contains(undesiredIngredient)) {
                history.addLine("..Result does not contain " + undesiredIngredient + ". So continues.");
                continue;
            }
            else if (history.getAdaptationPoints() >= adaptationThreshold) {
                history.addLine(".. Adaptation threshold: " + adaptationThreshold + " is reached. Can not consider: " + undesiredIngredient);
                undesiredIngredientsAreSatisfied = false;
                continue;
            }
            if (Operations.ingredientIsValid(undesiredIngredient)) {

                // consider simply remove rules
                if (this.considerRemoveRules(resultValues, undesiredIngredient)) {
                    simpleRemoveRuleFound = true;
                }
                // consider substitution rules
                else if (Operations.attributeContainsValues(recipeAttribute)) {

                    // consider substitution rules
                    if (this.considerSubstitutionRules(resultValues, undesiredIngredient)) {
                        substitutionRuleFound = true;
                    }
                    // consider similarity substitution rules
                    else {
                        if (this.considerSimilaritySubstitutions(resultValues, queryValues, attrDesc, undesiredIngredient)) {
                            similaritySubstitutionRuleFound = true;
                        }
                    }
                }
                else {
                    history.addLine("..No ingredients on that Attribute in recipe to substitute with");
                }

                // If some of the undesired ingredients are not satisfied, the whole adapted recipe should not be considered
                if (!substitutionRuleFound && !simpleRemoveRuleFound && !similaritySubstitutionRuleFound) {
                    history.addLine("The undesired ingredient " + undesiredIngredient + " is not satisfied");
                    undesiredIngredientsAreSatisfied = false;
                }
            }
            else {
                history.addLine("..QueryValue is invalid");
            }

            history.addLine("");
        }

        // If no values are added to resultList and it was previously unknown, return null (then it will be set to unknown again)
        return Operations.getListAsString(resultValues);
    }

    private Boolean considerRemoveRules(List<String> resultValues, String undesiredIngredient) {
        if (SubstitutionMethods.findRemoveRule(allCurrentIngredients, undesiredIngredient)) {
            history.addLine("..Remove rule found: Removes " + undesiredIngredient);
            resultValues.remove(undesiredIngredient);
            adaptationSubstitutions.add(new AdaptationSubstitution(undesiredIngredient, "X"));
            modificationIsMade = true;
            history.addUndesiredRemoveRuleCount();
            return true;
        }
        else {
            history.addLine("..No remove rule found");
            return false;
        }
    }

    private Boolean considerSubstitutionRules(List<String> resultValues, String undesiredIngredient) {
        HashSet<String> allUndesiredIngredients = Operations.getAllIngredientsFromInstance(queryUndesired);
        String substitutionIngredient = SubstitutionMethods.findRemovalSubstitution(undesiredIngredient, allUndesiredIngredients, allCurrentIngredients);
        if (substitutionIngredient != null) {
            history.addLine("..SubstitutionRule remove rule found");
            history.addLine(undesiredIngredient + " --> " + substitutionIngredient);
            resultValues.add(substitutionIngredient);
            resultValues.remove(undesiredIngredient);
            adaptationSubstitutions.add(new AdaptationSubstitution(undesiredIngredient, substitutionIngredient));
            modificationIsMade = true;
            history.addUndesiredSubstitutionRuleCount();

            // update substitution ingredient attribute
            AttributeDesc substitutionIngredientAttrDesc = Operations.findAttrDescForValue(substitutionIngredient);
            Attribute substitutionIngredientAttr = adaptedInstance.getAttForDesc(substitutionIngredientAttrDesc);
            List<String> substitutionIngredientAttrValues = Operations.getAttributeAsSingleValues(substitutionIngredientAttr);
            substitutionIngredientAttrValues.add(substitutionIngredient);
            String ModifiedAttributeOfferAsString = Operations.getListAsString(substitutionIngredientAttrValues);
            Operations.modifyAttrDesc(substitutionIngredientAttrDesc, ModifiedAttributeOfferAsString, history, adaptedInstance);

            return true;
        }
        else {
            history.addLine("..No substitution remove rule found");
            return false;
        }
    }

    private Boolean considerSimilaritySubstitutions(List<String> resultValues, List<String> queryValues, AttributeDesc attrDesc, String undesiredIngredient) throws Exception {
        history.addLine("- Consider similarity substitutions for " + undesiredIngredient);
        double threshold = Settings.getThresholdForIngredientSubstitution(attrDesc);
        double bestScore = 0.0;
        List<String> bestSubstitutionIngredients = new ArrayList<String>();

        try {
            List<AmalgamationFct> availableAmalgamFcts = CBREngine.getConcept().getAvailableAmalgamFcts();
            // Loop through each amalgamationFct (Taxonomies for attributeDescs)
            for (AmalgamationFct amFct : availableAmalgamFcts) {
                // Get taxonomy for current attrDesc
                TaxonomyFct activeFct = ((TaxonomyFct) amFct.getActiveFct(attrDesc));
                // Get nodes in that taxonomy
                List<TaxonomyNode> nodes = activeFct.getTopSymbol().getNodes();
                // Loop through nodes to find best substitution ingredient
                for (TaxonomyNode n : nodes) {
                    String node = n.toString();
                    // Local similarity between undesired ingredient and ingredient in taxonomy
                    Attribute undesiredAttribute = attrDesc.getAttribute(undesiredIngredient);
                    Attribute taxonomyAttribute = attrDesc.getAttribute(node);
                    double localSimilarity = CBREngine.getProject().getLocalSimilarity(undesiredAttribute, taxonomyAttribute, attrDesc).getRoundedValue();
                    // force to not substitute with synonyms (similarity 1)
                    // check that subst ing is not an undesired ingredient as well and that it is not undesired ingredient
                    if (!node.equals(undesiredIngredient) && !queryValues.contains(node) && localSimilarity != 1.0 && !resultValues.contains(node)) {
                        if (localSimilarity > bestScore) {
                            // reset bestSubstitutionIngredients and set only this ingredient there
                            bestSubstitutionIngredients.clear();
                            bestSubstitutionIngredients.add(node);
                            bestScore = localSimilarity;
                        }
                        else if (localSimilarity == bestScore) {
                            bestSubstitutionIngredients.add(node);
                        }
                    }
                }
            }
        } catch (Exception e) {
            System.out.println("");
            System.out.println(e);
            System.out.println("Error in considerSimilaritySubstitutions - undesired query");
        }

        if (bestSubstitutionIngredients.size() != 0 && bestScore > threshold) {
            history.addLine(".. chooses random among current best node(s): " + bestSubstitutionIngredients.toString());
            Random randomGenerator = new Random();
            int index = randomGenerator.nextInt(bestSubstitutionIngredients.size());
            String winner = bestSubstitutionIngredients.get(index);

            history.addLine(".. Found match: " + winner + ", score: " + bestScore);
            resultValues.add(winner);
            resultValues.remove(undesiredIngredient);
            adaptationSubstitutions.add(new AdaptationSubstitution(undesiredIngredient, winner));
            history.addUndesiredSimilaritySubstitutionCount();
            modificationIsMade = true;
            return true;
        }

        else {
            history.addLine(".. No similarity match satisfy the threshold");
            return false;
        }
    }

    public Boolean undesiredIngredientsAreSatisfied() { return undesiredIngredientsAreSatisfied; }
    public Boolean modificationIsMade() { return modificationIsMade; }
    public List<AdaptationSubstitution> getSubstitutions() { return adaptationSubstitutions; }

}
