package no.ntnu.mycbr.rest.adaptation.helpClasses;

import de.dfki.mycbr.core.casebase.Attribute;
import de.dfki.mycbr.core.model.AttributeDesc;
import de.dfki.mycbr.core.similarity.Similarity;
import no.ntnu.mycbr.CBREngine;
import no.ntnu.mycbr.rest.adaptation.adaptedInstance.AdaptationHistory;
import no.ntnu.mycbr.rest.adaptation.rules.AdaptationSubstitution;
import no.ntnu.mycbr.rest.adaptation.rules.SimpleRule;
import no.ntnu.mycbr.rest.adaptation.rules.SubstitutionRule;
import de.dfki.mycbr.core.casebase.Instance;

import java.util.*;

public class SubstitutionMethods {

    // returns substitution if found, otherwise null
    public static AdaptationSubstitution considerSimpleAdditionRules(List<String> resultValues,
                                                                     String desiredIngredient,
                                                                     HashSet<String> allCurrentIngredients,
                                                                     AdaptationHistory history) {
        for (SimpleRule simpleAddition : CBREngine.getRules().additionRules) {
            if (simpleAddition.requirementsAreSatisfied(allCurrentIngredients, desiredIngredient)) {
                resultValues.add(desiredIngredient);
                AdaptationSubstitution subst = new AdaptationSubstitution("X", desiredIngredient);
                history.addDesiredAdditionRuleCount();
                history.addLine(".. simply adding rule found: X -->" + desiredIngredient);
                return subst;
            }
        }
        history.addLine(".. no simply adding rule found");
        return null;
    }

    // returns substitution if found, otherwise null
    public static AdaptationSubstitution considerDesiredSubstitutionRules(Attribute queryAttribute,
                                                             List<String> resultValues,
                                                             String desiredIngredient,
                                                             HashSet<String> allCurrentIngredients,
                                                             AdaptationHistory history) {
        String offer = SubstitutionMethods.findSubstitutionOffer(queryAttribute, desiredIngredient, allCurrentIngredients);
        if (offer != null) {
            history.addLine(".. substitution rule found: " + offer + " --> " + desiredIngredient);
            resultValues.remove(offer);
            resultValues.add(desiredIngredient);
            AdaptationSubstitution subst = new AdaptationSubstitution(offer, desiredIngredient);
            history.addDesiredSubstitutionRuleCount();
            return subst;
        }
        else {
            history.addLine(".. no substitution rule found");
            return null;
        }
    }

    // returns substitution if found, otherwise null
    private static String findSubstitutionOffer(Attribute queryAttribute, String desiredIngredient, HashSet<String> ingredientAlternatives) {
        // we don't want to find an offer (substitute out) any desired ingredient or children of desired ingredients
        HashSet<String> queryValues = new HashSet<String>(Operations.getAttributeAsSingleValues(queryAttribute));
        HashSet<String> allDesiredAndTheirChildren = OperationsTaxonomy.getChildrenOfSeveralIngredients(queryValues);
        List<String> offerAlternatives = new ArrayList<String>();
        for (SubstitutionRule substitutionRule : CBREngine.getRules().substitutionRules) {
            String offer = substitutionRule.getSubstitutionOffer(desiredIngredient, ingredientAlternatives, allDesiredAndTheirChildren);
            if (offer!=null) {
                offerAlternatives.add(offer);
            }
        }
        // return random offer among offer alternatives
        if (offerAlternatives.size() != 0) {
            Random randomGenerator = new Random();
            int index = randomGenerator.nextInt(offerAlternatives.size());
            return offerAlternatives.get(index);
        }
        // if no offer alternatives, return null
        return null;
    }

    // returns substitution if found, otherwise null
    public static AdaptationSubstitution considerSimilaritySubstituion(AttributeDesc attrDesc,
                                                                 String desiredIngredient,
                                                                 List<String> resultValues,
                                                                 Instance queryDesired,
                                                                 AdaptationHistory history) throws Exception {
        HashSet<String> allIngredientsFromInstance = Operations.getAllIngredientsFromInstance(queryDesired);
        HashSet<String> allDesiredIngredientsAndChildren = OperationsTaxonomy.getChildrenOfSeveralIngredients(allIngredientsFromInstance);

        double bestScore = 0.0;
        String bestOriginalIngredient = null;
        AdaptationSubstitution subst = null;
        try {
            Attribute desiredIngAttribute = attrDesc.getAttribute(desiredIngredient);

            for (String offerAlternative : resultValues) {
                Attribute offerAlternativeAttribute = attrDesc.getAttribute(offerAlternative);
                Similarity localSimilarity = CBREngine.getProject().getLocalSimilarity(desiredIngAttribute, offerAlternativeAttribute, attrDesc);
                if (localSimilarity.getRoundedValue() > bestScore && !allDesiredIngredientsAndChildren.contains(offerAlternative) && localSimilarity.getRoundedValue() != 1.0) {
                    bestScore = localSimilarity.getRoundedValue();
                    bestOriginalIngredient = offerAlternative;
                }
            }
            if (bestOriginalIngredient != null && bestScore > Settings.getThresholdForIngredientSubstitution(attrDesc)) {
                resultValues.add(desiredIngredient);
                resultValues.remove(bestOriginalIngredient);
                subst = new AdaptationSubstitution(bestOriginalIngredient, desiredIngredient);
                history.addDesiredSimilaritySubstitutionCount();
                history.addLine(".. similarity substitution: " + bestOriginalIngredient + " --> " + desiredIngredient + " with sims: " + bestScore);
            }
            else {
                history.addLine(".. similarity substitution: " + bestOriginalIngredient + "; " + bestScore + " found, but does not satisfy similarity threshold");
            }

        } catch (Exception e) {
            System.out.println(e);
            System.out.println("Error in considerSimilaritySubstitution: Desired ingredient");
        }
        return subst;
    }


    // UNDESIRED SUBSTITUTIONS

    // returns substitution if found, otherwise null
    public static String findRemovalSubstitution(String undesiredIngredient, HashSet<String> allUndesiredIngredients, HashSet<String> allIngredients) {
        List<String> substitutionAlternatives = new ArrayList<String>();
        for (SubstitutionRule substitutionRule : CBREngine.getRules().substitutionRules) {
            String subst = substitutionRule.getRemovalSubstitution(undesiredIngredient, allIngredients);
            if (subst!=null && !allUndesiredIngredients.contains(subst)) {
                substitutionAlternatives.add(subst);
            }
        }
        // return random subst
        if (substitutionAlternatives.size() != 0) {
            Random randomGenerator = new Random();
            int index = randomGenerator.nextInt(substitutionAlternatives.size());
            return substitutionAlternatives.get(index);
        }
        // if no substitution alternatives, return null
        return null;
    }

    // returns true if a removal rule was found
    public static Boolean findRemoveRule(HashSet<String> allIngredients, String undesiredIngredient) {
        for (SimpleRule simpleDeletion : CBREngine.getRules().deletionRules) {
            if (simpleDeletion.requirementsAreSatisfied(allIngredients, undesiredIngredient)) {
                return true;
            }
        }
        return false;

    }

}
