package no.ntnu.mycbr;

import de.dfki.mycbr.core.Project;
import de.dfki.mycbr.core.casebase.Instance;
import de.dfki.mycbr.core.casebase.MultipleAttribute;
import de.dfki.mycbr.core.model.Concept;
import de.dfki.mycbr.core.model.SymbolDesc;
import de.dfki.mycbr.core.similarity.Similarity;
import de.dfki.mycbr.util.Pair;
import junit.framework.TestCase;
import no.ntnu.mycbr.rest.App;
import no.ntnu.mycbr.rest.NewRetrieval;
import no.ntnu.mycbr.rest.adaptation.PostAdaptationAdapted;
import no.ntnu.mycbr.rest.adaptation.PostAdaptationOriginal;
import no.ntnu.mycbr.rest.adaptation.PreAdaptation;
import no.ntnu.mycbr.rest.adaptation.adaptedInstance.AdaptationHistory;
import no.ntnu.mycbr.rest.adaptation.adaptedInstance.DesiredAdaptation;
import no.ntnu.mycbr.rest.adaptation.adaptedInstance.SuitableAdaptation;
import no.ntnu.mycbr.rest.adaptation.adaptedInstance.UndesiredAdaptation;
import no.ntnu.mycbr.rest.adaptation.helpClasses.Operations;
import no.ntnu.mycbr.rest.adaptation.rules.AdaptationSubstitution;
import no.ntnu.mycbr.rest.adaptation.rules.RulesInit;
import org.junit.Test;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class AdaptationTest extends TestCase {
    private Project testProject;
    private Concept testConcept;
    SymbolDesc vegetableDesc;
    SymbolDesc basicDesc;
    SymbolDesc fishDesc;
    SymbolDesc supplementDesc;
    Instance originalInstance;
    Instance adaptedInstance;
    Instance queryDesired;
    Instance queryUndesired;

    public void initTestProject() {

        CBREngine cbrEngine = new CBREngine();
        Project project = cbrEngine.createTestCBRProject();
        CBREngine.setConceptName("TestConcept");
        App.setProject(project);
        Concept concept = project.getConceptByID(CBREngine.getConceptName());
        CBREngine.setRules(new RulesInit());

        Collection<Instance> allInstances = concept.getAllInstances();
        this.testConcept = concept;
        this.testProject = project;

        this.vegetableDesc = (SymbolDesc) testConcept.getAttributeDesc("Vegetable");
        this.basicDesc = (SymbolDesc) testConcept.getAttributeDesc("Basic");
        this.fishDesc = (SymbolDesc) testConcept.getAttributeDesc("Fish");
        this.supplementDesc = (SymbolDesc) testConcept.getAttributeDesc("Supplement");

        originalInstance = new Instance(testConcept, "originalInstance");
        adaptedInstance = new Instance(testConcept, "adaptedInstance");
        queryDesired = new Instance(testConcept, "queryDesired");
        queryUndesired = new Instance(testConcept, "queryUndesired");
    }

    @Test
    public void testExtensionOfUndesiredQuery() {
        System.out.println(".. Run testExtensionOfUndesiredQuery");
        try {
            this.initTestProject();
            MultipleAttribute multipleAttributeQuery = HelpMethodsForTesting.getMultipleAttribute(vegetableDesc, "tomato");
            queryUndesired.addAttribute(vegetableDesc, multipleAttributeQuery);
            Instance resultInstance = PreAdaptation.addChildrenOfUndesiredIngredients(queryUndesired, testConcept);
            List<String> undesiredIngredients = Operations.getAttributeAsSingleValues(resultInstance.getAttForDesc(vegetableDesc));
            assertTrue("Length of undesired ingredients is: " + undesiredIngredients.size() + " but should be 3 ", undesiredIngredients.size() == 3);

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testAdditionSubst() {
        System.out.println(".. Run testAdditionSubst");
        try {
            this.initTestProject();

            // Required ingredients (in original recipe) for the simple addition rule to work
            MultipleAttribute multipleAttributeOriginal = HelpMethodsForTesting.getMultipleAttribute(vegetableDesc, "lettuce,tomato");
            originalInstance.addAttribute(vegetableDesc, multipleAttributeOriginal);
            adaptedInstance.addAttribute(vegetableDesc, multipleAttributeOriginal);

            // When these ingredients are in original recipe, cucumber should simple be added (query requirement)
            MultipleAttribute multipleAttributeQuery = HelpMethodsForTesting.getMultipleAttribute(vegetableDesc, "cucumber");
            queryDesired.addAttribute(vegetableDesc, multipleAttributeQuery);

            // Start desired adaptation
            AdaptationHistory history = new AdaptationHistory(originalInstance, adaptedInstance);
            DesiredAdaptation desiredAdaptation = new DesiredAdaptation(originalInstance, adaptedInstance, queryDesired, queryUndesired, history, 10, testConcept);

            // Check substitution results
            List<AdaptationSubstitution> substitutions = desiredAdaptation.getSubstitutions();

            assertTrue("Length of substitution list should be 1, but is: " + substitutions.size(), substitutions.size() == 1);
            assertTrue("Number of simple addition rules used in statistics should be 1, but is: " + history.getDesiredAdditionRuleCount(),
                    history.getDesiredAdditionRuleCount() == 1);
            AdaptationSubstitution subst = substitutions.get(0);
            assertTrue("Nothing should have been substituted out, but instead this ingredient was substiuted out: " + subst.getIngredient1(),
                    subst.getIngredient1().equals("X"));
            assertTrue("Cucumber should have been substited in, but instead this ingredient was substitted in: " + subst.getIngredient2(),
                    subst.getIngredient2().equals("cucumber"));

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testDesiredSubst() {
        System.out.println(".. Run testDesiredSubst");
        try {
            this.initTestProject();
            MultipleAttribute multipleAttributeOriginal = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "ketchup");
            originalInstance.addAttribute(supplementDesc, multipleAttributeOriginal);
            adaptedInstance.addAttribute(supplementDesc, multipleAttributeOriginal);

            // When these ingredients are in original recipe, cucumber should simple be added (query requirement)
            MultipleAttribute multipleAttributeQuery = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "tomato paste");
            queryDesired.addAttribute(supplementDesc, multipleAttributeQuery);

            // Start desired adaptation
            AdaptationHistory history = new AdaptationHistory(originalInstance, adaptedInstance);
            DesiredAdaptation desiredAdaptation = new DesiredAdaptation(originalInstance, adaptedInstance, queryDesired, queryUndesired, history, 10, testConcept);

            // Check substitution results
            List<AdaptationSubstitution> substitutions = desiredAdaptation.getSubstitutions();

            assertTrue("Length of substitution list should be 1, but is: " + substitutions.size(), substitutions.size() == 1);
            assertTrue("Number of desired substitution rules used in statistics should be 1, but is: " + history.getDesiredSubstitutionRuleCount(),
                    history.getDesiredSubstitutionRuleCount() == 1);
            AdaptationSubstitution subst = substitutions.get(0);
            assertTrue("Ketchup should have been substituted out, but instead this ingredient was substiuted out: " + subst.getIngredient1(),
                    subst.getIngredient1().equals("ketchup"));
            assertTrue("Tomato paste should have been substited in, but instead this ingredient was substitted in: " + subst.getIngredient2(),
                    subst.getIngredient2().equals("tomato paste"));


        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testDesiredSimilaritySubst() {
        System.out.println(".. Run testDesiredSimilaritySubst");
        try {
            this.initTestProject();


            MultipleAttribute multipleAttributeOriginal2 = HelpMethodsForTesting.getMultipleAttribute(fishDesc, "salmon");
            originalInstance.addAttribute(fishDesc, multipleAttributeOriginal2);
            adaptedInstance.addAttribute(fishDesc, multipleAttributeOriginal2);

            // When these ingredients are in original recipe, cucumber should simple be added (query requirement)
            MultipleAttribute multipleAttributeQuery = HelpMethodsForTesting.getMultipleAttribute(fishDesc, "cod");
            queryDesired.addAttribute(fishDesc, multipleAttributeQuery);

            // Start desired adaptation
            AdaptationHistory history = new AdaptationHistory(originalInstance, adaptedInstance);
            DesiredAdaptation desiredAdaptation = new DesiredAdaptation(originalInstance, adaptedInstance, queryDesired, queryUndesired, history, 10, testConcept);

            // Check substitution results
            List<AdaptationSubstitution> substitutions = desiredAdaptation.getSubstitutions();

            assertTrue("Length of substitution list should be 1, but is: " + substitutions.size(), substitutions.size() == 1);
            assertTrue("Number of desired substitution rules used in statistics should be 1, but is: " + history.getDesiredSimilaritySubstitutionCount(),
                    history.getDesiredSimilaritySubstitutionCount() == 1);
            AdaptationSubstitution subst = substitutions.get(0);
            assertTrue("Salmon should have been substituted out, but instead this ingredient was substiuted out: " + subst.getIngredient1(),
                    subst.getIngredient1().equals("salmon"));
            assertTrue("Cod should have been substited in, but instead this ingredient was substitted in: " + subst.getIngredient2(),
                    subst.getIngredient2().equals("cod"));


        } catch (Exception e) {
            e.printStackTrace();
        }

    }

    @Test
    public void testDeletionSubst() {
        System.out.println(".. Run testDeletionSubst");
        try {
            this.initTestProject();

            MultipleAttribute multipleAttributeOriginal = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "ketchup,tabasco sauce");
            originalInstance.addAttribute(supplementDesc, multipleAttributeOriginal);
            adaptedInstance.addAttribute(supplementDesc, multipleAttributeOriginal);

            MultipleAttribute multipleAttributeQuery = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "tabasco sauce");
            queryUndesired.addAttribute(supplementDesc, multipleAttributeQuery);

            // Start undesired adaptation
            AdaptationHistory history = new AdaptationHistory(originalInstance, adaptedInstance);
            UndesiredAdaptation undesiredAdaptation = new UndesiredAdaptation(originalInstance, adaptedInstance, queryUndesired, history, 10, testConcept);

            // Check substitution results
            List<AdaptationSubstitution> substitutions = undesiredAdaptation.getSubstitutions();

            assertTrue("Length of substitution list should be 1, but is: " + substitutions.size(), substitutions.size() == 1);
            assertTrue("Number of desired substitution rules used in statistics should be 1, but is: " + history.getUndesiredRemoveRuleCount(),
                    history.getUndesiredRemoveRuleCount() == 1);
            AdaptationSubstitution subst = substitutions.get(0);
            assertTrue("Tabasco sauce should have been substituted out, but instead this ingredient was substiuted out: " + subst.getIngredient1(),
                    subst.getIngredient1().equals("tabasco sauce"));
            assertTrue("Nothing should have been substited in, but instead this ingredient was substitted in: " + subst.getIngredient2(),
                    subst.getIngredient2().equals("X"));

        } catch (Exception e) {

        }
    }

    @Test
    public void testUndesiredSubst() {
        System.out.println(".. Run testUndesiredSubst");
        try {
            this.initTestProject();
            MultipleAttribute multipleAttributeOriginal = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "tomato paste");
            originalInstance.addAttribute(supplementDesc, multipleAttributeOriginal);
            adaptedInstance.addAttribute(supplementDesc, multipleAttributeOriginal);

            // When these ingredients are in original recipe, cucumber should simple be added (query requirement)
            MultipleAttribute multipleAttributeQuery = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "tomato paste");
            queryUndesired.addAttribute(supplementDesc, multipleAttributeQuery);

            // Start desired adaptation
            AdaptationHistory history = new AdaptationHistory(originalInstance, adaptedInstance);
            UndesiredAdaptation undesiredAdaptation = new UndesiredAdaptation(originalInstance, adaptedInstance, queryUndesired, history, 10, testConcept);

            // Check substitution results
            List<AdaptationSubstitution> substitutions = undesiredAdaptation.getSubstitutions();

            assertTrue("Length of substitution list should be 1, but is: " + substitutions.size(), substitutions.size() == 1);
            assertTrue("Number of desired substitution rules used in statistics should be 1, but is: " + history.getUndesiredSubstitutionRuleCount(),
                    history.getUndesiredSubstitutionRuleCount() == 1);
            AdaptationSubstitution subst = substitutions.get(0);
            assertTrue("Tomato paste should have been substituted out, but instead this ingredient was substiuted out: " + subst.getIngredient1(),
                    subst.getIngredient1().equals("tomato paste"));
            assertTrue("Ketchup paste should have been substited in, but instead this ingredient was substitted in: " + subst.getIngredient2(),
                    subst.getIngredient2().equals("ketchup"));


        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testUndesiredSimilaritySubst() {
        System.out.println(".. Run testUndesiredSimilaritySubst");
        try {
            this.initTestProject();
            MultipleAttribute multipleAttributeOriginal = HelpMethodsForTesting.getMultipleAttribute(fishDesc, "tuna");
            originalInstance.addAttribute(fishDesc, multipleAttributeOriginal);
            adaptedInstance.addAttribute(fishDesc, multipleAttributeOriginal);

            MultipleAttribute multipleAttributeQuery = HelpMethodsForTesting.getMultipleAttribute(fishDesc, "tuna");
            queryUndesired.addAttribute(fishDesc, multipleAttributeQuery);

            // Start desired adaptation
            AdaptationHistory history = new AdaptationHistory(originalInstance, adaptedInstance);
            UndesiredAdaptation undesiredAdaptation = new UndesiredAdaptation(originalInstance, adaptedInstance, queryUndesired, history, 10, testConcept);

            // Check substitution results
            List<AdaptationSubstitution> substitutions = undesiredAdaptation.getSubstitutions();

            assertTrue("Length of substitution list should be 1, but is: " + substitutions.size(), substitutions.size() == 1);
            assertTrue("Number of undesired similarity substitution used in statistics should be 1, but is: " + history.getUndesiredSimilaritySubstitutionCount(),
                    history.getUndesiredSimilaritySubstitutionCount() == 1);
            AdaptationSubstitution subst = substitutions.get(0);
            assertTrue("Tuna should have been substituted out, but instead this ingredient was substiuted out: " + subst.getIngredient1(),
                    subst.getIngredient1().equals("tuna"));
            assertTrue("Salmon or cod should have been substited in, but instead this ingredient was substituted in: " + subst.getIngredient2(),
                    subst.getIngredient2().equals("salmon") || subst.getIngredient2().equals("cod"));


        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testSuitableIngredients() {
        System.out.println(".. Run testSuitableIngredients");
        try {
            this.initTestProject();
//             Create original results mock
            MultipleAttribute multipleAttributeOriginal2 = HelpMethodsForTesting.getMultipleAttribute(fishDesc, "tuna,salmon");
            adaptedInstance.addAttribute(fishDesc, multipleAttributeOriginal2);
            MultipleAttribute multipleAttributeOriginal3 = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "ketchup");
            adaptedInstance.addAttribute(supplementDesc, multipleAttributeOriginal3);

            List<AdaptationSubstitution> adaptationSubstitutions = new ArrayList<AdaptationSubstitution>();
            adaptationSubstitutions.add(new AdaptationSubstitution("X", "salmon"));

            AdaptationHistory history = new AdaptationHistory();

            SuitableAdaptation.considerSuitableIngredients(history, adaptationSubstitutions, adaptedInstance, queryUndesired, queryDesired);

            List<String> attributeAsSingleValues = Operations.getAttributeAsSingleValues(adaptedInstance.getAttForDesc(supplementDesc));

            assertTrue("Supplement should now contain suitable ingredient salmon rillettes, but contains instead: " + attributeAsSingleValues.toString(), attributeAsSingleValues.contains("salmon rillettes"));

        } catch (Exception e) {
            e.printStackTrace();
        }

    }

    @Test
    public void testRetrieval() {
        try {
            this.initTestProject();
            MultipleAttribute multipleAttributeOriginal = HelpMethodsForTesting.getMultipleAttribute(fishDesc, "tuna");
            adaptedInstance.addAttribute(fishDesc, multipleAttributeOriginal);

            MultipleAttribute multipleAttributeQuery = HelpMethodsForTesting.getMultipleAttribute(fishDesc, "salmon");
            queryDesired.addAttribute(fishDesc, multipleAttributeQuery);

            Pair<Instance, Similarity> instanceSimilarityPair = NewRetrieval.retrieveSingle(queryDesired, adaptedInstance, this.testConcept);
            Double sim = instanceSimilarityPair.getSecond().getRoundedValue();
            assertTrue("Similarity score should be 0.8, but is: " + sim, sim == 0.8);


            MultipleAttribute multipleAttributeOriginal2 = HelpMethodsForTesting.getMultipleAttribute(vegetableDesc, "lettuce");
            adaptedInstance.addAttribute(vegetableDesc, multipleAttributeOriginal2);

            MultipleAttribute multipleAttributeQuery2 = HelpMethodsForTesting.getMultipleAttribute(vegetableDesc, "tomato");
            queryDesired.addAttribute(vegetableDesc, multipleAttributeQuery2);

            Pair<Instance, Similarity> instanceSimilarityPair2 = NewRetrieval.retrieveSingle(queryDesired, adaptedInstance, this.testConcept);
            Double sim2 = instanceSimilarityPair2.getSecond().getRoundedValue();
            assertTrue("Similarity score should be 0.8, but is: " + sim2, sim2 == 0.65);


        } catch (Exception e) {

        }
    }

    @Test
    public void testNoUndesiredProposed() {
        System.out.println(".. Run testNoUndesiredProposed");
        try {
            this.initTestProject();
            // undesired query
            MultipleAttribute multipleAttributeQuery = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "mayonnaise,mustard");
            queryUndesired.addAttribute(supplementDesc, multipleAttributeQuery);

            // create mocked retrieved results
            List<Pair<Instance, Similarity>> retrievedResults = new ArrayList<Pair<Instance, Similarity>>();

            // create instances to add
            Instance instance1 = new Instance(testConcept, "instance1");
            instance1.addAttribute(supplementDesc, HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "mayonnaise,ketchup"));
            Instance instance2 = new Instance(testConcept, "instance2");
            instance2.addAttribute(supplementDesc, HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "ketchup,pesto"));
            Instance instance3 = new Instance(testConcept, "instance3");
            instance3.addAttribute(supplementDesc, HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "pesto"));
            Instance instance4 = new Instance(testConcept, "instance4");
            instance4.addAttribute(supplementDesc, HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "ketchup,pesto,mayonnaise"));
            Instance instance5 = new Instance(testConcept, "instance5");
            instance5.addAttribute(supplementDesc, HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "ketchup,mustard"));

            // get pairs and add to retrieved results
            Pair<Instance, Similarity> pair1 = new Pair<Instance, Similarity>(instance1, Similarity.get(0.5));
            retrievedResults.add(pair1);
            Pair<Instance, Similarity> pair2 = new Pair<Instance, Similarity>(instance2, Similarity.get(0.5));
            retrievedResults.add(pair2);
            Pair<Instance, Similarity> pair3 = new Pair<Instance, Similarity>(instance3, Similarity.get(0.5));
            retrievedResults.add(pair3);
            Pair<Instance, Similarity> pair4 = new Pair<Instance, Similarity>(instance4, Similarity.get(0.5));
            retrievedResults.add(pair4);
            Pair<Instance, Similarity> pair5 = new Pair<Instance, Similarity>(instance5, Similarity.get(0.5));
            retrievedResults.add(pair5);

            // run filtering code
            List<Pair<Instance, Similarity>> modifiedRetrievedList = PostAdaptationOriginal.filterOriginalRecipes(testConcept, queryUndesired, retrievedResults);

            // should contain case 2, 3 because they does not contain mayonnaise
            assertTrue("Filtered result list should contain case 2 because it does not contains mayonnaise", modifiedRetrievedList.contains(pair2));
            assertTrue("Filtered result list should contain case 3 because it does not contains mayonnaise", modifiedRetrievedList.contains(pair3));

            // should not contain cases 5 because it contain mustard
            assertTrue("Filtered result list should NOT contain case 5 because it contains mustard", !modifiedRetrievedList.contains(pair5));

            // should not contain cases 1,4 because they contain mayonnaise
            assertTrue("Filtered result list should NOT contain case 1 because it contains mayonnaise", !modifiedRetrievedList.contains(pair1));
            assertTrue("Filtered result list should NOT contain case 4 because it contains mayonnaise", !modifiedRetrievedList.contains(pair4));

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testNoDuplicatesProposed() {
        System.out.println(".. Run testNoDuplicatesProposed");
        try {
            this.initTestProject();
            Instance instance1 = new Instance(testConcept, "originalInstance1");
            Instance instance2 = new Instance(testConcept, "originalInstance2");
            Instance instance3 = new Instance(testConcept, "originalInstance3");
            Instance instance4 = new Instance(testConcept, "originalInstance4");

            // Instance 1 and instance 2 are set to the same
            MultipleAttribute multipleAttribute1 = HelpMethodsForTesting.getMultipleAttribute(vegetableDesc, "lettuce,tomato");
            instance1.addAttribute(vegetableDesc, multipleAttribute1);
            instance2.addAttribute(vegetableDesc, multipleAttribute1);
            MultipleAttribute multipleAttribute2 = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "mayonnaise");
            instance1.addAttribute(supplementDesc, multipleAttribute2);
            instance2.addAttribute(supplementDesc, multipleAttribute2);

            // Instance 3 is unique
            MultipleAttribute multipleAttribute3 = HelpMethodsForTesting.getMultipleAttribute(vegetableDesc, "cucumber,lettuce");
            instance3.addAttribute(vegetableDesc, multipleAttribute3);
            MultipleAttribute multipleAttribute4 = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "guacamole");
            instance3.addAttribute(supplementDesc, multipleAttribute4);

            // Instance 4 is unique
            MultipleAttribute multipleAttribute5 = HelpMethodsForTesting.getMultipleAttribute(vegetableDesc, "lettuce,tomato");
            instance4.addAttribute(vegetableDesc, multipleAttribute5);
            MultipleAttribute multipleAttribute6 = HelpMethodsForTesting.getMultipleAttribute(supplementDesc, "tzatziki");
            instance4.addAttribute(supplementDesc, multipleAttribute6);

            // Add all instances to "result list"
            List<Pair<Instance, Similarity>> originalRecipes = new ArrayList<Pair<Instance, Similarity>>();
            List<Pair<Instance, Similarity>> adaptedRecipes = new ArrayList<Pair<Instance, Similarity>>();

            Pair<Instance, Similarity> pair1 = new Pair<Instance, Similarity>(instance1, Similarity.get(0.5));
            originalRecipes.add(pair1);
            Pair<Instance, Similarity> pair2 = new Pair<Instance, Similarity>(instance2, Similarity.get(0.5));
            adaptedRecipes.add(pair2);
            Pair<Instance, Similarity> pair3 = new Pair<Instance, Similarity>(instance3, Similarity.get(0.5));
            originalRecipes.add(pair3);
            Pair<Instance, Similarity> pair4 = new Pair<Instance, Similarity>(instance4, Similarity.get(0.5));
            adaptedRecipes.add(pair4);

            // run code
            PostAdaptationAdapted.removeEqualRecipes(originalRecipes, adaptedRecipes);

            assertTrue("Length of substitution list should be 1, but is: " + adaptedRecipes.size(), adaptedRecipes.size()==1);
            assertTrue("Instance2 should have been removed from unique list, but is not", !adaptedRecipes.contains(instance2));
            assertTrue("Instance4 should be in unique list", !adaptedRecipes.contains(instance4));
        } catch (Exception e) {
            e.printStackTrace();
        }

    }
}
