import {Injectable} from "@angular/core";
import {Http, RequestMethod, RequestOptions, Response, URLSearchParams} from "@angular/http";
import {Observable} from "rxjs/Observable";
import {AppSettings} from "../../appSettings";
import {Case, KeyValuePair, Recipe} from "./home.model";
import {PREPARATIONLINKS} from "./shared/preparationlinks";

import "rxjs/add/operator/map";
import "rxjs/add/operator/catch";

@Injectable()

export class HomeService {

  static shouldExcludeAttribute(attribute: string) {
    let excludedAttributes = AppSettings.EXCLUDED_ATTRIBUTES;
    for (let i = 0; i < excludedAttributes.length; i++) {
      if (attribute === excludedAttributes[i]) {
        return true;
      }
    }
    return false;
  }

  static getPreparationLinks() {
    return PREPARATIONLINKS;
  }

  private static handleError(error: any) {
    let errMsg = (error.message) ? error.message : error.status ? `${error.status} - ${error.statusText}` : 'Server error';
    console.error(errMsg);
    return Observable.throw(errMsg);
  }

  constructor(private http: Http) {
  }

  getAttributes(): Observable<KeyValuePair[]> {
    let params: URLSearchParams = new URLSearchParams();
    params.set('concept name', AppSettings.CONCEPT_NAME);

    return this.http.get(AppSettings.BACKEND_URL + '/attributes', {search: params})
      .map((response: Response) => response.json().attributes)
      .map(result => {
        let attributes: KeyValuePair[] = [];
        let keys = Object.keys(result);

        for (let i = 0; i < keys.length; i++) {
          if (!HomeService.shouldExcludeAttribute(keys[i])) {
            attributes.push(new KeyValuePair(keys[i], result[keys[i]]));
          }
        }
        return attributes;
      })
      .catch(HomeService.handleError);
  }

  getValuesForAttribute(attribute: string): Observable<string[]> {
    let params: URLSearchParams = new URLSearchParams();
    params.set('concept name', AppSettings.CONCEPT_NAME);
    params.set('attribute name', attribute);

    return this.http.get(AppSettings.BACKEND_URL + '/values', {search: params})
      .map((response: Response) => response.json().valueRange)
      .map(result => {
        return result[attribute];
      })
      .catch(HomeService.handleError);
  }

  getSimilarCases(caseBaseName: string, desiredValues: string, undesiredValues: string, doAdaptation: boolean): Observable<Case[]> {
    let params: URLSearchParams = new URLSearchParams();
    params.set('casebase', caseBaseName);
    params.set('concept name', AppSettings.CONCEPT_NAME);

    let options = new RequestOptions({
      method: RequestMethod.Post,
      search: params
    });

    let queryContent = '{ "desired": ' + desiredValues + ', "undesired": ' + undesiredValues + '}';
    let body: JSON = JSON.parse(queryContent);

    let callString: string = doAdaptation === true ? '/adaptation' : '/retrieval';

    return this.http.post(AppSettings.BACKEND_URL + callString, body, options)
      .map((response: Response) => response.json().similarCases)
      .map((result) => {
          let cases: Case[] = [];
          let caseIndex = Object.keys(result);
          for (let i = 0; i < caseIndex.length; i++) {
            let similarCase = new Case(result[caseIndex[i]].caseName, result[caseIndex[i]].score);

            // Extract possible oldIngredient-newIngredient pairs
            let retrievedAdaptations: KeyValuePair[] = [];
            if (result[caseIndex[i]].adaptations === null || result[caseIndex[i]].adaptations.length === 0) {
              retrievedAdaptations = [];
              similarCase.adapted = false;
            } else {
              similarCase.adapted = true;
              for (let object of result[caseIndex[i]].adaptations) {
                retrievedAdaptations.push(new KeyValuePair(object.first, object.second));
              }
            }
            similarCase.adaptations = retrievedAdaptations;

            // Extract possible parentInQuery-childInRecipe pairs
            let queryParentIngredientChildList: KeyValuePair[] = [];
            if (result[caseIndex[i]].parentChildren === null) {
              queryParentIngredientChildList = [];
            } else {
              for (let object of result[caseIndex[i]].parentChildren) {
                queryParentIngredientChildList.push(new KeyValuePair(object.first, object.second));
              }
            }
            similarCase.queryParentIngredientChildList = queryParentIngredientChildList;

            // Extract possible childInQuery-parentInRecipe pairs
            let queryChildIngredientParentList: KeyValuePair[] = [];
            if (result[caseIndex[i]].childrenParent === null) {
              queryChildIngredientParentList = [];
            } else {
              for (let object of result[caseIndex[i]].childrenParent) {
                queryChildIngredientParentList.push(new KeyValuePair(object.first, object.second));
              }
            }
            similarCase.queryChildIngredientParentList = queryChildIngredientParentList;

            cases.push(similarCase);
          }
          return cases;
        }
      )
      .catch(HomeService.handleError);
  }

  getCase(casebaseName: string, caseId: string): Observable<KeyValuePair[]> {
    let params: URLSearchParams = new URLSearchParams();
    params.set('casebase', casebaseName);
    params.set('concept name', AppSettings.CONCEPT_NAME);
    params.set('caseID', caseId);

    return this.http.get(AppSettings.BACKEND_URL + '/case', {search: params})
      .map((response: Response) => response.json().case)
      .map((result) => {
        let attributeList: KeyValuePair[] = [];
        let keys = Object.keys(result);
        for (let i = 0; i < keys.length; i++) {
          attributeList.push(new KeyValuePair(keys[i], result[keys[i]]));
        }
        return attributeList;
      })
      .catch(HomeService.handleError);
  }

  addCase(casebaseName: string, adaptedCaseAttributeString: string, adaptedRecipeString: string) {
    let params: URLSearchParams = new URLSearchParams();
    params.set('casebase', casebaseName);
    params.set('concept name', AppSettings.CONCEPT_NAME);
    params.set('recipe', adaptedRecipeString);

    let options = new RequestOptions({
      method: RequestMethod.Post,
      search: params
    });

    let body: JSON = JSON.parse(adaptedCaseAttributeString);

    return this.http.post(AppSettings.BACKEND_URL + '/addcase', body, options)
      .map((response: Response) => {
        return response;
      })
      .catch(HomeService.handleError);
  }

  getRecipe(recipeId: string): Observable<Recipe> {
    let params: URLSearchParams = new URLSearchParams();
    params.set('recipeId', recipeId);

    return this.http.get(AppSettings.BACKEND_URL + '/recipe', {search: params})
      .map((response: Response) => response.json().recipe)
      .map((result) => {
        let recipe: Recipe = new Recipe();
        recipe.id = result.id;
        recipe.createdByAI = result.createdByAI;
        recipe.ingredients = [];
        recipe.preparationsteps = [];
        for (let ingredient of result.ingredients) {
          recipe.ingredients.push(new KeyValuePair(ingredient.model, ingredient.display));
        }
        for (let step of result.preparationsteps) {
          recipe.preparationsteps.push(step);
        }
        return recipe;
      })
      .catch(HomeService.handleError);
  }
}
